"""sous-chef - A small webapp for viewing and searching Chef nodes"""

from __future__ import absolute_import

__version__ = '1.0.1'
__author__ = 'Sam Clements <sam.clements@datasift.com>'
__all__ = ['create_app', 'create_debug_app']


from sous_chef.app import create_app, create_debug_app



"""A wrapper around PyChef to deal with configuration"""

from __future__ import absolute_import

import collections
import urllib  # urllib.parse for Python 3

import chef


class FlaskChefAPI(chef.ChefAPI):
    """Used to build a API client from a Flask app's configuration"""

    @classmethod
    def configure(cls, app):
        keys = ['CHEF_URL', 'CHEF_KEY', 'CHEF_CLIENT']
        args = [app.config[k] for k in keys]
        return cls(*args) if all(args) else chef.autoconfigure()


class PartialSearch(collections.Iterable):
    """Chef partial search implemenation, using PyChef

    A class similar to the existing Search class in PyChef, using Chef's
    partial search API - this only returns the requested keys and is much
    faster than a full search.

    Chef Search documentation:
        http://docs.opscode.com/essentials_search.html
        http://docs.opscode.com/api_chef_server_search_index.html
    """

    url = '/search'

    def __init__(self, index, q='*:*', keys={}, rows=1000, start=0, api=None):
        self.name = index
        self.keys = keys
        self.api = api or chef.api.ChefAPI.get_global()

        self._args = dict(q=q, rows=rows, start=start)
        self.url = '/{url}/{index}?{args}'.format(
            url=self.__class__.url,
            index=index,
            args=urllib.urlencode(dict(q=q, rows=rows, start=start)))

    @staticmethod
    def keys_from_object(keys):
        # Coerce lists of dotted attibutes to a dict
        if isinstance(keys, list):
            keys = [key.split('.') for key in keys]
            keys = dict((key[0], key) for key in keys)

        # Always return the name and chef environment
        keys.setdefault('name', ['name'])
        keys.setdefault('chef_environment', ['chef_environment'])

        return keys

    @property
    def body(self):
        return self.keys_from_object(self.keys)

    @property
    def data(self):
        if not hasattr(self, '_data'):
            self._data = self.api.api_request('POST', self.url, data=self.body)
        return self._data

    def __iter__(self):
        for row in self.data['rows']:
            yield row['data']

    def __len__(self):
        return len(self.data['rows'])

    def __getitem__(self, value):
        return self.data['rows'][value]


def get_node(name, keys):
    """Returns a single node using Chef partial search"""
    nodes = PartialSearch('node', 'name:' + name, keys=keys)
    if not nodes:
        raise chef.exceptions.ChefError('No node found')
    elif len(nodes) > 1:
        raise chef.exceptions.ChefError('Too many results returned')
    else:
        return nodes[0]['data']



"""The create_app() function"""

from __future__ import absolute_import

import flask

import sous_chef.blueprints
import sous_chef.chef

__all__ = ['create_app']


def create_app():
    app = flask.Flask('sous_chef', instance_relative_config=True)

    # Load configuration from defaults and an optional instance config file
    app.config.from_object('sous_chef.defaults')
    app.config.from_pyfile('config.py')

    # Configure chef before the first request
    app.chef = sous_chef.chef.FlaskChefAPI.configure(app)

    # Register blueprints - currently only the user interface
    app.register_blueprint(sous_chef.blueprints.ui)

    return app


def create_debug_app():
    app = create_app()
    app.config['DEBUG'] = True

    # Install the Flask debug toolbar extension, if the package is availible
    try:
        from flask_debugtoolbar import DebugToolbarExtension
    except ImportError:
        pass
    else:
        app.config['DEBUG_TB_INTERCEPT_REDIRECTS'] = False
        DebugToolbarExtension(app)

    return app



"""Flask blueprints for sous-chef"""

from __future__ import absolute_import

import flask
import chef
import chef.exceptions

from sous_chef.chef import PartialSearch, get_node

__all__ = ['ui']


ui = flask.Blueprint('ui', __name__)


@ui.before_request
def set_chef_api_client():
    """Set the global ChefAPI object as the default for this thread"""
    flask.current_app.chef.set_default()


# Nodes

@ui.route('/', endpoint='home')
@ui.route('/nodes/')
def node_index():
    nodes = PartialSearch('node', keys=['name', 'chef_environment'])
    return flask.render_template('node_index.html', nodes=nodes)


@ui.route('/nodes/<string:name>')
def node(name):
    return flask.render_template('node.html', node=get_node(name, [
        'chef_environment',
        'roles',
        'run_list',
        'packages'
    ]))


# Roles

@ui.route('/roles/')
def role_index():
    return flask.render_template('role_index.html', roles=chef.Role.list())


@ui.route('/roles/<string:name>')
def role(name):
    return flask.render_template(
        'role.html',
        role=chef.Role(name),
        nodes=PartialSearch('node', 'roles:' + name))


# Environments

@ui.route('/environments/')
def environment_index():
    return flask.render_template(
        'environment_index.html', environments=chef.Environment.list())


@ui.route('/environments/<string:name>')
def environment(name):
    return flask.render_template(
        'environment.html',
        environment=chef.Environment(name),
        nodes=PartialSearch('node', 'chef_environment:' + name))


# Packages

@ui.route('/packages/<string:type>/<string:name>')
def package(type, name):
    nodes = PartialSearch('node', 'packages_{0}:{1}'.format(type, name), keys={
        'package_version': ['packages', type, name, 'version']
    })
    return flask.render_template(
        'package.html', package_type=type, package_name=name, nodes=nodes)


# def current_envionment():
#     if not 'chef_environment' in flask.session:
#         flask.session['chef_environment'] = flask.current_app.config.get(
#             'DEFAULT_CHEF_ENVIRONMENT', '_default')
#     return flask.session['chef_environment']

# @ui.before_request
# def set_environment_variables():
#     """Ensure a list of environments is availible"""
#     flask.g.chef_environments = sorted(chef.Environment.list())
#     flask.g.chef_environment = current_envionment()

# @ui.route('/environments/<string:name>/select')
# def select_environment(name):
#     if name in flask.g.chef_environments:
#         flask.session['chef_environment'] = name
#         flask.session.permanent = True
#     return flask.redirect(flask.request.referrer or flask.url_for('ui.home'))
